﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Security;
using System.Threading;

namespace NativeWithSpan
{
    // C# 7.2 - Span<T>
    // http://www.sysnet.pe.kr/2/0/11534

    // .NET Core 3.0
    class Program
    {
        public Program()
        {
        }

        static unsafe void Main(string[] args)
        {
            int bufSize = 40960;

            Action<int> a1 = (count) =>
            {
                while (count-- > 0)
                {
                    int[] buf = new int[bufSize];
                    buf[0] = 0;
                    buf[bufSize - 1] = 0;
                }
            };

            Action<int> a2 = (count) =>
            {
                while (count-- > 0)
                {
                    using (NativeMemory<int> buf = new NativeMemory<int>(bufSize))
                    {
                        Span<int> viewBuf = buf.GetView();
                        viewBuf[0] = 0;
                        viewBuf[bufSize - 1] = 0;
                    }
                }
            };

            Action<int, Action<int>> perfTest = (count, action) =>
            {
                Stopwatch st = new Stopwatch();
                st.Start();
                action(count);
                st.Stop();

                Console.WriteLine(st.ElapsedMilliseconds);
            };

            perfTest(1, a1);
            perfTest(1, a2);

            checkGCFunc();
            perfTest(1000000, a1);
            checkGCFunc();

            Console.WriteLine();

            checkGCFunc();
            perfTest(1000000, a2);
            checkGCFunc();
        }

        static int checkCount = 0;
        private static void checkGCFunc()
        {
            int count = GC.CollectionCount(2);
            // Console.WriteLine($"{checkCount++} : {count}");
        }
    }

    public unsafe ref struct NativeMemory<T> where T : unmanaged
    {
        int _size;
        IntPtr _ptr;

        public NativeMemory(int size)
        {
            _size = size;

            long lSize = _size;
            lSize *= sizeof(T);
                
            IntPtr bufSize = new IntPtr(lSize);
            _ptr = Marshal.AllocHGlobal(bufSize);
        }

        public Span<T> GetView()
        {
            return new Span<T>(_ptr.ToPointer(), _size);
        }

        // C# 8.0에서만 using과 함께 사용 가능
        public void Dispose()
        {
            if (_ptr == IntPtr.Zero)
            {
                return;
            }

            Marshal.FreeHGlobal(_ptr);
            _ptr = IntPtr.Zero;
        }
    }
}
